// PresAction.hpp

#ifndef presaction_hpp
#define presaction_hpp

#include <vector>

namespace Prescribble {

class Note;

/**
 * PresAction 
 * - a potentially undoable/redoable action in a Note
 */
    class Action {
    public:
        
        struct Context;
        
        virtual ~Action(){}
        
        virtual bool undo( Context *a_pContext = 0 ) = 0;
        virtual bool redo( Context *a_pContext = 0 ) = 0;
        virtual bool isDone() = 0;
    };
    typedef Action PresAction;
    
    
    class ActionBase : public Action {
    protected:
        Note *m_pNote;
        bool m_done;
    public:
        ActionBase( Note *a_pNote )
            : m_pNote( a_pNote ), m_done(false)
            {}
        virtual bool undo( Context *a_pContext = 0 ){return false;}
        virtual bool redo( Context *a_pContext = 0 ){return false;}
        virtual bool isDone(){return m_done;}
    };
    typedef ActionBase PresActionBase;
    
    
/**
 * Paramter Object pattern for the actions.
 */
    struct Action::Context {
    };
    
    class ComplexAction : public ActionBase
    {
        typedef std::vector<Action*> ActionList;
    protected:
        ActionList m_subactions;
    public:
        
        ComplexAction() : ActionBase(0){}
        virtual ~ComplexAction();
        
        virtual bool undo( Context *a_pContext = 0 );
        virtual bool redo( Context *a_pContext = 0 );
        void add( Action& );
        
        bool empty(){ return m_subactions.empty(); }
        
        int size(){ return m_subactions.size(); }
    };
    typedef ComplexAction PresComplexAction;
    
}; // namespace Prescribble

#endif
